	SECTION .text
HLINE	;; ( c x1 x2 y -- , Draw a horizontal line )
	mpush	eax, ecx, edx, edi
	dpop	eax
	mul	dword [gr.w]
	dpop	ecx
	sub	ecx, S(0)
	inc	ecx
	mov	edi, [gr.buf]	;top left
	add	edi, eax	;	move down
	add	edi, S(0)	;	more right
	drop
	dpop	eax		;color
	cld
	rep	stosb
	mpop	eax, ecx, edx, edi
	ret
eHLINE


VLINE	;; ( c x y1 y2 -- , Draw a vertical line )
	mpush	eax, ecx, edx, edi
	dpop	ecx
	dpop	eax
	sub	ecx, eax	;length
	inc	ecx
	mul	dword [gr.w]	;move down
	add	eax, S(0)	;move right
	drop
	add	eax, [gr.buf]
	mov	edi, eax
	dpop	eax		;color
	mov	edx, [gr.w]
.plot	mov	[edi], al	;plot pixel	*** critical loop ***
	add	edi, edx	;next line	(much slower than HLINE)
	loop	.plot
	mpop	eax, ecx, edx, edi
	ret
eVLINE


PUTPIXEL	;; ( c x y -- )
	mpush	eax, edx, edi
	dpop	eax		;y
	mul	dword [gr.w]
	dpop	edi		;x
	add	edi, eax
	add	edi, [gr.buf]
	dpop	eax		;c
	stosb
	mpop	eax, edx, edi
	ret
ePUTPIXEL


BLIT	;; ( w h x1 y1 x2 y2 -- , copy a rectangular area )
	mpush	eax, ebx, ecx, edx, esi, edi
	dpop	eax		;dest
	mul	dword [gr.w]
	add	eax, S(0)
	drop
	add	eax, [gr.buf]
	mov	edi, eax
	dpop	eax		;src
	mul	dword [gr.w]
	add	eax, S(0)
	drop
	add	eax, [gr.buf]
	mov	esi, eax
	dpop	ebx		;w
	dpop	ecx		;h
	mov	edx, [gr.w]
	sub	edx, ebx
.1	push	ecx
	mov	ecx, ebx
	rep	movsb
	pop	ecx
	add	esi, edx
	add	edi, edx
	loop	.1
	mpop	eax, ebx, ecx, edx, esi, edi
	ret
eBLIT


%if 0
LINE            ; ( c x1 y1 x2 y2 -- , Draw a diagonal line )
;
;Algorithm:
; Use fixed-point numbers (16.16 in 32-bit, 10.6 in 16-bit)
; Start at topmost point
; Calculate dx (dy=1)
; Count = |y2-y1| (vertical distance)
; (x,y) = initial point (in fixed-point notation)
; Loop:
;   hline from x to x+dx (convert to integer)
;   x = x+dx
;   y++
;
	mpush	eax,ebx,ecx,edx,esi,edi

        mov     ax, [vid.addr]
        mov     es, ax

        call    ROT             ;( c x1 x2 y2 y1 )
        upop    eax             ;y1
        upop    ecx             ;y2

        ; Find topmost point    AX = initial y-value
        cmp     eax, ecx
        jb      .noswap

        xchg    eax, ecx
        call    SWAP            ;( c x2 x1 )
.noswap

        inc     ecx
        sub     ecx, eax        ;CX = count = vertical distance

        mul     cell [vid.w]
        mov     esi, eax        ;SI = initial scanline (y * screen-width)

        upop    eax             ;Final x-value
        inc     eax
        upop    edi             ;Initial x-value

        sub     eax, edi        ;Calculate dx (in fixed-point)
        xor     edx, edx
        shl     eax, FIXSHIFT
        idiv    ecx
        mov     ebx, eax        ;BX = dx (delta-x)

        mov     edx, edi
        shl     edx, FIXSHIFT   ;DX = current x-value (fixed-point)

        upop    eax             ;AL = color

        cmp     edx, 0
        jge     .positive_dx
        std     ;If negative, EDI must be decremented during REP STOSB
.positive_dx

        add     edi, esi        ;DI = current point address
.loop	push    ecx             ;Draw hline
        ; This next piece may seem unnecessary; maybe it is. It's for accuracy.
        mov     ecx, edx        ; Get x
        add     edx, ebx        ; Add dx
        sub     edx, ecx        ; Find difference
        xchg    edx, ecx
        shr     ecx, FIXSHIFT
        rep     stosb
        pop     ecx

        add     edi, [vid.w]
        loop    .loop

	mpop	eax,ebx,ecx,edx,esi,edi
	ret
eLINE
%endif


%if 0
CIRCLE          ; ( c x y r -- , Draw a circle, radius=r at x,y )
;
; Variables:
;  cx  = DI
;  cy  = SI
;  df  = DX
;  de  = AX
;  dse = CX
;  temp = BX (for pixel plotting)
;  x = S1
;  y = S0
;  c = S2
;
; Optimize: Precalculate address (for cy, at least?)
;
	mpush	eax,ebx,ecx,edx,esi,edi

        mov     ax, [vid.addr]
        mov     es, ax

        ;Initialize variables
        sub     di, di          ;cx = 0
        upop    si              ;cy = radius
        mov     ax, 3           ;de = 0

        mov     dx, si
        neg     dx
        inc     dx              ;df = 1-r = -r+1

        mov     cx, si
        sal     cx, 1           ;dse = 2r
        neg     cx
        add     cx, 5           ;dse = -2r+5

.loop
        ;Plot 8 points (1 per octant of circle)
        push    eax
        push    ecx

;Note: This is currently a text-mode circle plot :-)
;
        mov     ebx, S1         ;X
        add     ebx, edi
        mov     ecx, S0         ;Y
        add     ecx, esi
        upsh    ebx
        upsh    ecx
        call    AT_XY
        upsh    cell '*'
        call    EMIT

        mov     ebx, S1         ;X
        sub     ebx, edi
        mov     ecx, S0         ;Y
        add     ecx, esi
        upsh    ebx
        upsh    ecx
        call    AT_XY
        upsh    cell '*'
        call    EMIT

        pop     ecx
        pop     eax

        ;Update variables
        cmp     dx, 0           ;if (df < 0)
        jge     .gt
.lt	add     dx, ax          ;df+=de
        add     cx, 2           ;dse+=2
        jmp     .endif
.gt	add     dx, cx          ;df+=dse
        add     cx, 4           ;dse+=4
        dec     si              ;cy--
.endif	add     ax, 2           ;de+=2
        inc     di              ;cx++

        cmp     di, si          ;Repeat loop while (cx <= cy)
        jle near .loop

	mpop	eax,ebx,ecx,edx,esi,edi
	ret
eCIRCLE
%endif


	SECTION .data
gr
.buf	dd	0xA0000
.w	dd	320
.h	dd	200
